/**
	Provides access to the browser's DOM api: the
	document object, elements, and events.

	The document object also includes things like cookies,
	local storage, and a bunch of other random stuff.


	Note: this core file makes no attempt at cross-browser
	quirks. What you write here is pretty much directly
	sent to the browser in JS form.

	(You can test for the existence of an object at runtime
	by using the address of operator:

	if(&document.querySelector) {
		// go ahead and use it
	}
*/
module browser.document;

extern(js):
pragma(no_js_output):

import jslang.global;

JSDocument document;

// These can be reused in different classes
mixin template ElementNavigationFunctions(ElementType) {
	extern(js):
	pragma(no_js_output):
	ElementType getElementById(string);
	ElementType querySelector(string);
	ElementType[] querySelectorAll(string);
}

mixin template ElementNavigationProperties(ElementType) {
	extern(js):
	pragma(no_js_output):
	ElementType nextSibling;
	ElementType previousSibling;
	ElementType parentNode;

	ElementType offsetParent;
}

mixin template DocumentNavigationProperties(ElementType) {
	extern(js):
	pragma(no_js_output):
	ElementType documentElement;
	ElementType __js_body; // alas, body is a keyword in D, so have to do it this way. This refers to document.body directly in Javascript though.
	alias __js_body mainBody;
}

final class JSDocument {
	mixin ElementNavigationFunctions!(JSElement);
	mixin DocumentNavigationProperties!(JSElement);

	JSElement createElement(string);
	JSElement createTextNode(string);
	JSEvent createEvent(...);

	void write(string);

	string cookie;

	void addEventListener(string event, void delegate(JSEvent) handler, bool useCapture);
}

final class JSElement {
	string className;
	string tagName;

	string textContent; // not in old IE

	string innerHTML;
	alias innerHTML innerHtml; // I like having camel case available; this Just Works in D

	int nodeType;
	string nodeValue;

	mixin ElementNavigationFunctions!(JSElement);

	string getAttribute(string) const;
	void setAttribute(string, string);
	bool hasAttribute(string) const;

	mixin ElementNavigationProperties!(JSElement);

	JSElement appendChild(JSElement);
	void removeChild(JSElement);

	int offsetWidth;
	int offsetHeight;
	int offsetLeft;
	int offsetTop;

	int scrollLeft;
	int scrollTop;
	int scrollWidth;
	int scrollHeight;

	void addEventListener(string event, void delegate(JSEvent) handler, bool useCapture);
	void addEventListener(string event, void delegate() handler, bool useCapture);

	void addEventListener(string event, void function() handler, bool useCapture);

	JSCanvasContext getContext(string);
	JSStyle style;

	// Dynamic dataset; // this is p. useless here. just use get/set attribute

	/* only valid for specific tags */
	string href;

		/* images */
	string src;
	int width;
	int height;
	void delegate() onerror;
		/* form elements */
	string value;
	int selectedIndex;
	JSElement[] options;
	bool delegate() onsubmit;
		/* html5 video+audio */
	void play();
	void pause();
	float volume; // 0-1
	int currentTime; // to seek
	JSTimeRange seekable;
	JSTimeRange played;
}

final class JSTimeRange {
	int start(int item = 0);
	int end(int item = 0);
	int length;
}

final class JSStyle {
	string cssText;
	alias cssText this; // this is free and it is cool

	// gah I so want to make this dynamic but with zero wrapping..
	string display;
	string cssFloat;

	string clear;

	string listStyle;
	string listStyleType;
	string listStylePosition;
	string listStyleImage;

	string cursor;
	string clip;
	string overflow;

	string verticalAlign;
	string visibility;

	string height;
	string width;
	string maxHeight;
	string maxWidth;
	string minHeight;
	string minWidth;

	string position;
	string left;
	string right;
	string top;
	string bottom;
	string zIndex;

	string direction;
	string quotes;
	string univodeBidi;
	string font;
	string fontFamily;
	string fontStyle;
	string fontSize;
	string fontSizeAdjust;
	string fontVariant;
	string fontWeight;

	string letterSpacing;
	string lineHeight;

	string color;

	string textAlign;
	string textDecoration;
	string textShadow;
	string textIndent;
	string textTransform;
	string whiteSpace;
	string wordSpacing;

	string background;
	string backgroundAttachment;
	string backgroundColor;
	string backgroundPosition;
	string backgroundImage;
	string backgroundRepeat;

	string padding;
	string paddingLeft;
	string paddingTop;
	string paddingRight;
	string paddingBottom;

	string margin;
	string marginLeft;
	string marginTop;
	string marginRight;
	string marginBottom;

	string border;
	string borderColor;
	string borderStyle;
	string borderWidth;

	string borderLeft;
	string borderLeftColor;
	string borderLeftStyle;
	string borderLeftWidth;

	string borderRight;
	string borderRightColor;
	string borderRightStyle;
	string borderRightWidth;

	string borderTop;
	string borderTopColor;
	string borderTopStyle;
	string borderTopWidth;

	string borderBottom;
	string borderBottomColor;
	string borderBottomStyle;
	string borderBottomWidth;

	string outline;
	string outlineColor;
	string outlineStyle;
	string outlineWidth;

	string content;
	string counterIncrement;
	string counterReset;

	string orphans;
	string pageBreakAfter;
	string pageBreakBefore;
	string pageBreakInside;
	string widows;

	string borderCollapse;
	string borderSpacing;
	string captionSide;
	string emptyCells;
	string tableLayout;
}

final class Image {
	void delegate() onload;
	void delegate() onerror;
	string src;
	int width;
	int height;
}

final class JSCanvasContext {
	string fillStyle;
	string strokeStyle;
	float globalAlpha;
	float lineWidth;
	string lineCap;
	string lineJoin;

	float shadowOffsetX;
	float shadowOffsetY;
	float shadowBlur;
	string shadowColor;
	string font;

	string globalCompositeOperation;
	void clip();

	void fillText(string, float, float);
	void strokeText(string, float, float);

	void save();
	void restore();
	void translate(float, float);
	void rotate(float);
	void scale(float, float);
	void transform(float, float, float, float, float, float);
	void setTransform(float, float, float, float, float, float);

	void fillRect(float, float, float, float);
	void clearRect(float, float, float, float);
	void strokeRect(float, float, float, float);
	void beginPath();
	void closePath();
	void moveTo(float, float);
	void lineTo(float, float);
	void fill();
	void stroke();
	void arc(float, float, float, float, float, bool);
	void quadraticCurveTo(float, float, float, float);
	void bezierCurveTo(float, float, float, float, float, float);

	void rect(float, float, float, float);

	void drawImage(Image, float, float);
	void drawImage(Image, float, float, float, float, float, float, float, float);

	JSGradient createLinearGradient(float, float, float, float);
	JSGradient createRadialGradient(float, float, float, float, float, float);

	string createPattern(Image, string); // FIXME: doesn't *really* return string
}

final class JSGradient {
	string _d_hack;
	alias _d_hack this; // so we can assign it to fillStyle and strokeStyle

	void addColorStop(float, string);
}

final class JSEvent {
	string type;
	bool bubbles;
	bool cancelable;
	JSEvent currentTarget; // aka this in JS i think
	long timeStamp;

	void preventDefault();
	void stopPropagation();

	JSElement target;
	alias target srcElement;

	JSElement relatedTarget;

	int x;
	int y;
	int pageX;
	int pageY;
	int offsetX; // i don't think these are correct on firefox :-(
	int offsetY;
	int clientX;
	int clientY;
	int screenX;
	int screenY;

	bool ctrlKey;
	bool altKey;
	bool shiftKey;
	bool metaKey;
	
	int button;
	int which;
	int key;

	int keyCode;

	int buttons;

	// touchscreen
	JSTouch[] touches;

	// Apple additions
	JSTouch[] targetTouches;
	JSTouch[] changedTouches;

	float scale;
	float rotation;
}

final class JSTouch {
	int pageX;
	int pageY;
}

/*
class JSInputEvent : JSEvent {

}
*/

/// Just use character literals for many keys. A-Z, 0-0 space, enter, tab, esc, backspace
/// These are the other ones the browsers agree upon.
/// Use it in keydown and keyup events. which || keyCode in JS.
enum KeyCode {
	Shift = 16,
	Control = 17,
	Alt = 18,
	CapsLock = 20,
	NumLock = 144,

	LeftArrow = 37,
	UpArrow = 38,
	RightArrow = 39,
	DownArrow = 40,

	Insert = 45,
	Delete = 46,
	Home = 36,
	End = 35,
	PageUp = 33,
	PageDown = 34,

	F1 = 112,
	F2 = 113,
	F3 = 114,
	F4 = 115,
	F5 = 116,
	F6 = 117,
	F7 = 118,
	F8 = 119,
	F9 = 120,
	F10 = 121,
	F11 = 122,
	F12 = 123,
}
